@@ -1,8 +1,8 @@
 // Use all the g++ headerfiles
 
 // $Author: bothner $
-// $Revision: 1.9 $
-// $Date: 1992/05/14 22:09:28 $
+// $Revision: 1.11 $
+// $Date: 1993/06/06 20:28:06 $
 
 #include <_G_config.h>
 // If we have the old iostream library, it defines _OLD_STREAMS
@@ -72,7 +72,6 @@
 #include <Weibull.h>
 
 #include <assert.h>
-#include <bstring.h>
 #include <libc.h>
 #include <compare.h>
 #include <complex.h>
@@ -80,7 +79,6 @@
 #include <errno.h>
 #include <generic.h>
 #include <grp.h>
-#include <istream.h>
 #include <getpagesize.h>
 #include <time.h>
 #include <math.h>
@@ -88,7 +86,6 @@
 #include <minmax.h>
 #include <new.h>
 #include <osfcn.h>
-#include <ostream.h>
 #include <pwd.h>
 #include <regex.h>
 #include <setjmp.h>
@@ -98,11 +95,15 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <strclass.h>
-#include <streambuf.h>
 #include <string.h>
 #include <swap.h>
 #include <unistd.h>
 #include <values.h>
+#ifdef _IO_MAGIC
+#include <istream.h>
+#include <streambuf.h>
+#include <ostream.h>
+#endif
 
 main()
 {
