@@ -126,7 +126,7 @@ the executable file might be covered by the GNU General Public License. */
 #define FLT_ROUNDS 1
 #define FLT_RADIX 2
 #define DBL_DIG 15
-#define DBL_MAX_10_EXP (-307)
+#define DBL_MAX_10_EXP 308
 #define DBL_MAX_EXP 1024
 #endif
 
